/********************************************************************/
/*                        Ordinal numbers                           */
/********************************************************************/

#ifndef __ORD
#ifndef __HEADER__
#define __HEADER__
#define __ORD 1
#else
#define __ORD 0
#endif  // __HEADER__

#ifndef __cplusplus
#error Must use C++ for the type Ord.
#endif

#include <stdio.h>
#include <assert.h>
#include <alloc.h>
#include <values.h>

/************************** interface *******************************/

#define COMPARISON(_class, _op) \
		friend int operator _op (const _class &x, const _class &y) \
			{ return x.Comp(y) _op 0; }

#define COMPARISON_SET(_class) \
		COMPARISON (_class, ==)  \
		COMPARISON (_class, !=)  \
		COMPARISON (_class, <)   \
		COMPARISON (_class, <=)  \
		COMPARISON (_class, >)   \
		COMPARISON (_class, >=)


enum  // ordinal constants and operators (inner representation)
{
	// ordinal constants
	V_FIRST_CONST = -100,
	V_OMEGA,
	V_EPSILON,
	V_OMEGA1,
	V_LAST_CONST,

	// ordinal operators
	V_FIRST_BINOP = -200,
	V_PLUS, V_MINUS, V_TIMES, V_DIV, V_MOD, V_POWER,
	V_LIM, V_VAR,
	V_LAST_BINOP,

	// symbols for lexical analyzer
	V_FIRST_LEX = -300,
	V_LEFT_PARTH, V_RIGHT_PARTH, V_ENDSYMB, V_BADSYMB,
	V_LAST_LEX,

};  // enum

static char *OrdConstNames [] =
{
	"omega",
	"epsilon",
	"omega1",
};


// max. number of members calculated by CantOrd-power operator
#define ORD_POW_MAX_MEMB 100


// Ordinal Error handling:

extern int OrdError;  // Ord error status

enum // ordinal error codes
{
	ORD_ERR_OK = 0,
	ORD_ERR_SYNT,     // syntax analyzer error
	ORD_ERR_NAT,      // natural number calculator error (overflow etc.)
	ORD_ERR_DOM,      // ordinal domain error
	ORD_ERR_TOO,      // too many Cantor members in (temporary) result
	ORD_ERR_IMPL,     // feature not implemented yet
	ORD_ERR_OTHER,
};

#define SetOrdError(_val)  (OrdError = OrdError ? OrdError : (_val))



// natural number operators
#define NOPER(_op) \
		static long N##_op (long x, long y);
	NOPER (Plus)
	NOPER (Minus)
	NOPER (Times)
	NOPER (Div)
	NOPER (Mod)
	NOPER (Power)
#undef NOPER

/************** classes Ord, CantOrd declaration ********************/


class Ord  // Ordinal expression
{
	long value;
	Ord *left, *right;

	void Init ();
	Ord (long _value, Ord *_left, Ord *_right);

	// syntax analysis functions
	long GetSymb  (const char *&ptr) const;  // lexical analyzer
	Ord *SyntAn10 (const char *&ptr) const;  // factor +, -
	Ord *SyntAn20 (const char *&ptr) const;  // factor *, /, %
	Ord *SyntAn30 (const char *&ptr) const;  // factor ^
	Ord *SyntAn40 (const char *&ptr) const;  // atom or ()

public:

	// constructors and destructors
	Ord *Copy () const;
	Ord () { Init (); }
	Ord (long _value);
	Ord (const Ord & another);
	~Ord ();

	// operators overloading
	Ord & operator = (Ord & another);          // destructive assignment
	Ord & operator <<= (const Ord & another);  // non-destructive assignment

	int Comp (const Ord & another) const;
	COMPARISON_SET (Ord)

	#define ASSIGN(_op, _opval) \
		Ord & operator _op##= (const Ord &another) \
		{ return *this = Ord (_opval, Copy (), another.Copy ()); }
	ASSIGN (+, V_PLUS)
	ASSIGN (-, V_MINUS)
	ASSIGN (*, V_TIMES)
	ASSIGN (/, V_DIV)
	ASSIGN (%, V_MOD)
	ASSIGN (^, V_POWER)
	#undef ASSIGN

	#define ARITHM(_name, _op) \
		Ord & _name (const Ord &x, const Ord &y) \
		{ *this <<= x; return *this _op##= y; }
	ARITHM (Plus,  +)
	ARITHM (Minus, -)
	ARITHM (Times, *)
	ARITHM (Div,   /)
	ARITHM (Mod,   %)
	ARITHM (Power, ^)
	#undef ARITHM

	// manipulative functions
	long Value () const { return value; }
	Ord *GetLeft  () const { return left -> Copy (); }
	Ord *GetRight () const { return right -> Copy (); }
	Ord & Cantorize ();
	Ord & atoOrd (const char *expr);
	void PrintOn (FILE *stream) const;
	void PrintOn (FILE *stream, short priority) const;
	friend void OrdDebug ();  // for debugging purposes

}; // Ord

#define OMEGA    (Ord (V_OMEGA))
#define EPSILON  (Ord (V_EPSILON))
#define OMEGA1   (Ord (V_OMEGA1))


class CantOrd  // Ordinal numbers in Cantor normal form
{
	int      base;
	CantOrd *exp;
	long     mult;
	CantOrd *next;

	void Init ();

	#define GETOP(_name) \
		CantOrd *Get##_name (const CantOrd &x, const CantOrd &y) const;
	GETOP (Plus)
	GETOP (Minus)
	GETOP (Times)
	GETOP (Div)
	GETOP (Mod)
	GETOP (Power)
	#undef GETOP

	CantOrd & Renorm ();            // renorm (eg w^w^e = e, w+0 = w)
	CantOrd *GetRenormed () const;  // create renormed copy of *this
	CantOrd & Append (const CantOrd &another);
	CantOrd (long _base, CantOrd *_exp, long _mult, CantOrd *_next);

public:

	// constructors and destructors
	CantOrd *Copy () const;
	CantOrd () { Init (); }
	CantOrd (const CantOrd & another);
	CantOrd (const CantOrd & another, int member);
	CantOrd (const CantOrd & another, int memb_form, int memb_to);
	CantOrd (const Ord & ord_expr);
	~CantOrd ();

	// operators overloading
	CantOrd & operator = (CantOrd & another);         // destruct. assignment
	CantOrd & operator <<= (const CantOrd & another); // non-destr. assignment

	int Comp (const CantOrd & another) const;
	COMPARISON_SET (CantOrd)

	#define OPER(_name) \
		CantOrd & _name (const CantOrd &x, const CantOrd &y) \
		{                                                    \
			CantOrd *result = Get##_name (x, y);             \
			*this = *result;                                 \
			delete result;                                   \
			return *this;                                    \
		}
	OPER (Plus)
	OPER (Minus)
	OPER (Times)
	OPER (Div)
	OPER (Mod)
	OPER (Power)
	#undef OPER

	#define ASSIGN(_op, _proc) \
		CantOrd & operator _op##= (const CantOrd &another) \
		{ _proc (*this, another); return *this; }
	ASSIGN (+, Plus)
	ASSIGN (-, Minus)
	ASSIGN (*, Times)
	ASSIGN (/, Div)
	ASSIGN (%, Mod)
	ASSIGN (^, Power)
	#undef ASSIGN

	// math functions
	CantOrd *GetOrder () const;
     long GetMult () const {return mult;};

	// manipulative functions
	int NumMemb () const;
	Ord *ToOrd () const;

	// functions for debugging purposes:
	friend void CantOrdDebug ();    // to be implemented in test pgm
	void PrintOn (FILE *stream) const;
	void PrintOn (FILE *stream, short level) const;


}; // CantOrd

/************************** implementation **************************/

#if __ORD


int OrdError = 0;  // Ord error status global variable



/************ natural numbers' operators ****************************/


#define ASSERTXY  assert (x >= 0 && y >= 0);
#define NAT_ERROR(_result) \
	{ SetOrdError (ORD_ERR_NAT); return _result;}


static long NPlus (long x, long y)
{
	ASSERTXY
	if (MAXLONG - y < x) NAT_ERROR (MAXLONG)
	return x + y;
}

static long NMinus (long x, long y)
{
	ASSERTXY
	if (x < y) NAT_ERROR (x)
	return x - y;
}

static long NTimes (long x, long y)
{
	ASSERTXY
	if (x == 0 || y == 0) return 0;
	if (MAXLONG / y < x) NAT_ERROR (MAXLONG)
	return x * y;
}

static long NDiv (long x, long y)
{
	ASSERTXY
	if (y == 0) NAT_ERROR (x)
	return x / y;
}

static long NMod (long x, long y)
{
	ASSERTXY
	if (y == 0) NAT_ERROR (x)
	return x % y;
}

static long NPower (long x, long y)
{
	ASSERTXY
	if (x == 0) return y == 0 ? 1 : 0;
	if (x == 1) return 1;
	long result = 1;
	for (int i = 0; !OrdError && i < y; ++i) 
		// supposing MAXLONG==2^32, the loop iterates max.32 times:
		result = NTimes (result, x);
	return result;
}

#undef NAT_ERROR
#undef ASSERTXY

/***************** class Ord methods ********************************/


inline void Ord::Init ()
{ 
	value = 0; 
	left = right = NULL; 
}



inline Ord *Ord::Copy () const
{
	return this ? new Ord (value, left -> Copy (),  right -> Copy ())
		: NULL;
}


inline Ord::Ord (long _value, Ord *_left, Ord *_right)
{ 
	value = _value; 
	left = _left; 
	right = _right;
}



Ord::~Ord ()
{
	if (left)  delete (left);
	if (right) delete (right);
	Init ();
}


Ord & Ord::operator = (Ord & another)  // destructive assignment
{
	if (this != &another)
	{
		if (left)  delete (left);
		if (right) delete (right);
		value = another.value;
		left  = another.left;
		right = another.right;
		another.Init ();
	}
	return *this;
}


Ord & Ord::operator <<= (const Ord & another)  // non-destructive assign.
{
	if (this != &another)
	{
		Ord *copy = another.Copy ();
		*this = *copy;
		delete copy;
	}
	return *this;
}


Ord::Ord (const Ord & another)
{
	if (this != &another)
	{
		Init ();
		*this <<= another;
	}
}


inline Ord::Ord (long _value)
{
	assert (_value >= 0 || _value > V_FIRST_CONST && _value < V_LAST_CONST);
	value = _value;
	left = right = NULL;
}


Ord & Ord::Cantorize ()
{
	Ord *cantorized = CantOrd (*this).ToOrd ();
	*this = *cantorized;
	delete cantorized;
	return *this;
}


int Ord::Comp (const Ord &another) const
{
	return CantOrd (*this).Comp (CantOrd (another));
}


/****************** Ord input/output ********************************/


static long getsymb; // set by GetSymb


long Ord::GetSymb (const char *&ptr) const  // lexical analyzer
{
	// skip spaces
	while (*ptr == ' ') ++ptr;

	// symbolic name
	if (*ptr >= 'A' && *ptr <= 'Z' || *ptr >= 'a' && *ptr <= 'z')
	{
		const char *end = ptr;
		for (; 
			*end >= 'a' && *end <= 'z' ||
			*end >= 'A' && *end <= 'Z' ||
			*end >= '0' && *end <= '9'; 
			++end);
		for (long try = 0; try < V_LAST_CONST - V_FIRST_CONST - 1; ++try)
		{
			int found = 1;
			for (const char *p = OrdConstNames [try]; found && *p; ++p)
				found = *p == *(ptr + (p - OrdConstNames [try]));
			if (found && p - OrdConstNames [try] == end - ptr)
			{
				ptr = end;
				return getsymb = V_FIRST_CONST + try + 1;
			}
		}
		return getsymb = V_BADSYMB; // invalid symbolic name
	}

	// numeric value
	else if (*ptr >= '0' && *ptr <= '9')
	{	
		long value;
		for (value = 0; *ptr >= '0' && *ptr <= '9'; ++ptr)
			value = NPlus (NTimes (10, value), (*ptr - '0'));
		return getsymb = value;
	}

	// another character
	else switch (*(ptr++))
	{
		case '(':  return getsymb = V_LEFT_PARTH;
		case ')':  return getsymb = V_RIGHT_PARTH;
		case '+':  return getsymb = V_PLUS;
		case '-':  return getsymb = V_MINUS;
		case '*':  return getsymb = V_TIMES;
		case '/':  return getsymb = V_DIV;
		case '%':  return getsymb = V_MOD;
		case '^':  return getsymb = V_POWER;
		case '\0': return getsymb = V_ENDSYMB;
		default:   return getsymb = V_BADSYMB;
	}
}


Ord *Ord::SyntAn10 (const char *&ptr) const  // factor of +, -
{
	Ord *factor, *result;
	long op;

	for (result = SyntAn20 (ptr), op = getsymb;
		op == V_PLUS || op == V_MINUS;
		op = getsymb)
	{
		factor = (GetSymb (ptr), SyntAn20 (ptr));
		switch (op)
		{
			case V_PLUS:  *result += *factor; break;
			case V_MINUS: *result -= *factor; break;
			default: assert (0);
		}
		delete factor;
	}
	if (op != V_ENDSYMB &&
	    op != V_RIGHT_PARTH) SetOrdError (ORD_ERR_SYNT);
	return result;
}


Ord *Ord::SyntAn20 (const char *&ptr) const  // factor of *, /, %
{
	Ord *factor, *result;
	long op;

	for (result = SyntAn30 (ptr), op = getsymb;
		op == V_TIMES || op == V_DIV || op == V_MOD;
		op = getsymb)
	{
		factor = (GetSymb (ptr), SyntAn30 (ptr));
		switch (op)
		{
			case V_TIMES: *result *= *factor; break;
			case V_DIV:   *result /= *factor; break;
			case V_MOD:   *result %= *factor; break;
			default: assert (0);
		}
		delete factor;
	}
	return result;
}


Ord *Ord::SyntAn30 (const char *&ptr) const  // factor of ^
{
	Ord *factor, *result;
	long op;

	for (result = SyntAn40 (ptr), op = getsymb; op == V_POWER; op = getsymb)
	{
		factor = (GetSymb (ptr), SyntAn40 (ptr));
		switch (op)
		{
			case V_POWER: *result ^= *factor; break;
			default: assert (0);
		}
		delete factor;
	}
	return result;
}


Ord *Ord::SyntAn40 (const char *&ptr) const  // atom or ()
{
	Ord *result;

	// expression in parentheses
	if (getsymb == V_LEFT_PARTH)
	{
		result = (GetSymb (ptr), SyntAn10 (ptr));
		if (getsymb != V_RIGHT_PARTH) SetOrdError (ORD_ERR_SYNT);
		else GetSymb (ptr);
	}

	// atom
	else if (getsymb >= 0 || getsymb > V_FIRST_CONST && getsymb < V_LAST_CONST)
	{
		result = Ord (getsymb).Copy ();
		GetSymb (ptr);
	}

	// invalid symbol
	else
	{
		SetOrdError (ORD_ERR_SYNT);
		result = Ord ().Copy ();
		GetSymb (ptr);
	}
	return result;
}


Ord & Ord::atoOrd (const char *expr)
{
	Ord *result = (GetSymb (expr), SyntAn10 (expr));
	*this = *result;
	delete result;
	return *this;
}



void Ord::PrintOn (FILE *stream, short priority) const
{
	if (value >= 0)
	{ // numeric value
		fprintf (stream, "%ld", value);
		return;
	}

	if (value > V_FIRST_CONST && value < V_LAST_CONST)
	{ // ordinal constant
		fputs (OrdConstNames [(short) value - V_FIRST_CONST - 1], stream);
		return;
	}

	if (value > V_FIRST_BINOP && value < V_LAST_BINOP)
	{ // binary operator

		static struct OpDesc
		{
			const char *name;
			short priority, left_priority, right_priority;
		}
		opList[] =
		{
			{" + ", 10, 10, 10},
			{" - ", 10, 10, 11},
			{" * ", 20, 20, 20},
			{" / ", 20, 20, 21},
			{" % ", 20, 20, 21},
			{"^", 30, 30, 31},
		};
		const OpDesc *op = &opList [(short) value - V_FIRST_BINOP - 1];

		if (priority > op -> priority) fputc ('(', stream);
		left -> PrintOn (stream, op -> left_priority);
		fprintf (stream, "%s", op -> name);
		right -> PrintOn (stream, op -> right_priority);
		if (priority > op -> priority) fputc (')', stream);
		return;
	} // binary operator

	assert (0);  // invalid value
}

void Ord::PrintOn (FILE *stream) const
{
	PrintOn (stream, 0);
}


/********************* class CantOrd methods ************************/

inline void CantOrd::Init ()
{
	base = mult = 0;
	exp = next = NULL;
}


inline CantOrd *CantOrd::Copy () const
{
	return this ? new CantOrd (base, exp -> Copy (), mult, next -> Copy ())
		: NULL;
}


CantOrd::CantOrd (long _base, CantOrd *_exp, long _mult, CantOrd *_next)
{ 
	base = _base; 
	exp = _exp; 
	mult = _mult; 
	next = _next; 
}



CantOrd::~CantOrd ()
{
	if (next) delete next;
	if (exp)  delete exp;
	Init ();
}


CantOrd & CantOrd::operator = (CantOrd & another) // destructive assignment
{
	if (this != &another)
	{
		if (exp)  delete exp;
		if (next) delete next;
		base = another.base;
		exp  = another.exp;
		mult = another.mult;
		next = another.next;
		another.Init ();
	}
	return *this;
}


CantOrd & CantOrd::operator <<= (const CantOrd & another)  // non-destr.ass.
{
	if (this != &another)
	{
		CantOrd *copy = another.Copy ();
		*this = *copy;
		delete copy;
	}
	return *this;
}


CantOrd::CantOrd (const CantOrd &another)
{
	if (this != &another)
	{
		Init ();
		*this <<= another;
	}
}


CantOrd::CantOrd (const CantOrd &another, int member)
{
	Init ();
	if (member < 1) return;
	const CantOrd *ptr;
	for (ptr = &another; --member > 0 && ptr; ptr = ptr -> next);
	if (ptr)
		*this = CantOrd (ptr -> base, ptr -> exp -> Copy (), ptr -> mult, NULL);
}


CantOrd::CantOrd (const CantOrd &another, int n1, int n2)
{
	Init ();
	int num = another.NumMemb ();
	if (n2 < n1 || n1 > num || n2 < 1) return;
	if (n1 < 1)                        n1 = 1;
	if (n2 > num)                      n2 = num;

	CantOrd *result, *ptrr;
	const CantOrd *ptrc;

	for (ptrc = &another; --n2, --n1 > 0; ptrc = ptrc -> next);
	result = ptrc -> Copy ();
	for (ptrr = result; n2-- > 0; ptrr = ptrr -> next);
	delete ptrr -> next;
	ptrr -> next = NULL;
	*this = *result;
	delete result;
}


CantOrd::CantOrd (const Ord & ord_expr)
{
	Init ();
	long value = ord_expr.Value ();

	if (value == 0) return;

	if (value > 0)
	{
		*this = CantOrd (V_OMEGA, CantOrd (0).Copy (), value, NULL);
		return;
	}

	if (value == V_OMEGA)
	{
		*this = CantOrd (V_OMEGA, CantOrd (1).Copy (), 1, NULL);
		return;
	}

	if (value > V_FIRST_CONST && value < V_LAST_CONST)
	{
		*this = CantOrd (value, NULL, 1, NULL);
		return;
	}

	if (value > V_FIRST_BINOP && value < V_LAST_BINOP)
	{
		Ord *left  = ord_expr.GetLeft ();
		Ord *right = ord_expr.GetRight ();
		switch (value)
		{
			#define DEAL(_opval, _fnc) \
				case _opval: _fnc (CantOrd (*left), CantOrd(*right)); break;
			DEAL (V_PLUS,  Plus)
			DEAL (V_MINUS, Minus)
			DEAL (V_TIMES, Times)
			DEAL (V_DIV,   Div)
			DEAL (V_MOD,   Mod)
			DEAL (V_POWER, Power)
			#undef DEAL
			default: assert (0); // wrong operator
		}
		delete left;
		delete right;
		return;
	}
	assert (0); // invalid ordinal
}


int CantOrd::NumMemb () const  // number of members
{
	int num = 0;
	for (const CantOrd *ptr = this; ptr; ptr = ptr -> next) ++num;
	return num;
}


CantOrd & CantOrd::Append (const CantOrd &another)
{
	CantOrd *ptr;
	for (ptr = this; ptr -> next; ptr = ptr -> next);
	ptr -> next = another.Copy ();
	return *this;
}


CantOrd & CantOrd::Renorm ()
{
	CantOrd *result = GetRenormed ();
	*this = *result;
	delete result;
	return *this;
}


CantOrd *CantOrd::GetRenormed () const
{
	CantOrd *result;

	if (!this) return NULL;

	if (mult == 0)
		if (next == NULL)  // only 1 zero member
			return CantOrd (0).Copy ();
		else
			return next -> GetRenormed ();

	if (base == V_OMEGA)
		if (exp -> base != V_OMEGA
			&& exp -> base != 0
			&& exp -> next == NULL
			&& mult == 1)
		{
			// w^eps * 1 = eps
			result = CantOrd (exp -> base, NULL, 1, NULL).Copy ();
		}
		else
		{	// renorm exponent
			result = CantOrd (V_OMEGA, exp -> GetRenormed (), mult, NULL).Copy ();
		}
	else result = CantOrd (base, NULL, 1, NULL).Copy ();

	if (!next) return result;
	if (next -> mult == 0)
		// a zero-member follows: skip
		result -> next = next -> next -> GetRenormed ();
	else
		result -> next = next -> GetRenormed ();

	return result;
}


Ord *CantOrd::ToOrd () const
{
	Ord *result;
	do
	{
		if (base != V_OMEGA)
		{ // 0 or eps-number
			assert (exp == NULL); // cannot have exp
			result = Ord (base).Copy ();
			break;
		}
		assert (exp); // else must have exponent

		if (exp -> base == 0)
		{ // w^0
			assert (exp -> next == NULL); // cannot have exp -> next
			assert (next == NULL);        // cannot have next (renormed!)
			result = Ord (mult).Copy ();
			break;
		}

		if (exp -> base != V_OMEGA && exp -> next == NULL)
		{ // w^eps * n
			result = Ord (exp -> base).Copy ();
			if (mult != 1) *result *= mult;
			break;
		}

		result = OMEGA.Copy (); // else omega-based member

		if (exp -> base == V_OMEGA)
		{ // w^(w^x*n)*m
			assert (exp -> exp); // exp must have exponent

			if (exp -> exp -> base == 0)
			{ // w^(w^0 * n)
				assert (exp -> exp -> next == NULL); // exp cannot have next
				assert (exp -> next == NULL); // cannot have next (renorm'd!)
				assert (exp -> mult != 0);    // supposing renormed!

				if (exp -> mult == 1)
					if (mult != 1) *result *= mult;
					else;
				else
				{
					*result ^= exp -> mult;
					if (mult != 1) *result *= mult;
				}
				break;
			}
		}

		// else irreducible member
		Ord *temp = exp -> ToOrd ();
		if (mult == 1) *result ^= *temp;
		else         { *result ^= *temp; *result *= mult; }
		delete temp;

	} while (0);  // end Ord-izing this member

	if (next)
	{
		Ord *temp = next -> ToOrd ();
		*result += *temp;
		delete temp;
	}
	return result;
}


void CantOrd::PrintOn (FILE *stream, short level) const
{
	if (level);
	if (base != 0)
		fprintf (stream, "%s", OrdConstNames [(short) base - V_FIRST_CONST - 1]);
	else fprintf (stream, "0");
	if (exp)
	{
		fputc ('^', stream);
		if (exp -> next || exp -> base == V_OMEGA) fputc ('(', stream);
		exp -> PrintOn (stream, 1);
		if (exp -> next || exp -> base == V_OMEGA) fputc (')', stream);
	}
	if (base == V_OMEGA) fprintf (stream, " * %ld", mult);
	if (next)
	{
		fprintf (stream, " + ");
		next -> PrintOn (stream, 1);
	}
}


void CantOrd::PrintOn (FILE *stream) const
{
	PrintOn (stream, 0);
}




/****************** CantOrd mathematical functions ******************/


#define SIGN(x)	((x) > 0 ? 1 : ((x) < 0 ? -1 : 0))
#define TRUE 1
#define FALSE 0


int CantOrd::Comp (const CantOrd & another) const
{
	if (base != V_OMEGA && another.base != V_OMEGA)
	{ // both being ordinal constants
		long val1 = base         == 0 ? V_FIRST_CONST : base;
		long val2 = another.base == 0 ? V_FIRST_CONST : another.base;
		if (val1 != val2) return SIGN (val1 - val2);
		else goto Equal;
	}

	if (base == 0) return -1;         // *this is 0
	if (another.base == 0) return 1;  // another is 0

	if (base != V_OMEGA)              // *this (only) is an eps-number
		return CantOrd (V_OMEGA, CantOrd (base).Copy (), 1, next -> Copy ()).Comp (another);
	if (another.base != V_OMEGA)      // another (only) is an eps-number
		return Comp (CantOrd (V_OMEGA, CantOrd (another.base).Copy (), 1, another.next -> Copy ()));

	// both being ordinary members
	int comp = exp -> Comp (*another.exp);
	if (comp) return comp;
	comp = SIGN (mult - another.mult);
	if (comp) return comp;

Equal:;	// these members are equal - result depends on the next ones
	if (next == NULL && another.next == NULL) return 0;
	if (next == NULL) return -1;          // supposing renormed!
	if (another.next == NULL) return 1;
	return next -> Comp (*another.next);
}


CantOrd *CantOrd::GetOrder () const
{
	if (base == V_OMEGA) return exp -> Copy ();
	else return CantOrd (base).Copy ();
}


CantOrd *CantOrd::GetPlus (const CantOrd &x, const CantOrd &y) const
{
	CantOrd *result = CantOrd (0).Copy ();
	CantOrd *y_order = y.GetOrder ();
	const CantOrd *ptr = &x;
	int is_higher = TRUE;
	do
	{	// copy to result all members of x of higher order than y
		CantOrd *x_order = ptr -> GetOrder ();
		if (*x_order > *y_order)
		{
			result -> Append (CantOrd (*ptr, 1));
			ptr = ptr -> next;
		}
		else is_higher = FALSE;
		delete x_order;
	}
	while (ptr && is_higher);

	if (ptr)
	{
		CantOrd *x_order = ptr -> GetOrder ();
		if (*x_order == *y_order)
		{	// members of the same order
			result -> Append (CantOrd (V_OMEGA, x_order -> Copy (), NPlus (ptr -> mult, y.mult), NULL));
			if (y.next) result -> Append (*y.next);
		}
		else result -> Append (y);
		delete x_order;
	}
	else result -> Append (y);

	result -> Renorm ();
	delete y_order;
	return result;
}


CantOrd *CantOrd::GetMinus (const CantOrd &x, const CantOrd &y) const
{
	if (x < y)
	{
		SetOrdError (ORD_ERR_DOM);
		return x.Copy ();
	}

	CantOrd *result = CantOrd (0).Copy ();
	const CantOrd *ptrx = &x;
	const CantOrd *ptry = &y;
	int are_equal = TRUE;

	do  // skip all equal members
	{
		if (CantOrd (*ptrx, 1) == CantOrd (*ptry, 1))
		{
			ptrx = ptrx -> next;
			ptry = ptry -> next;
		}
		else are_equal = FALSE;
	}
	while (ptry && are_equal);

	if (ptrx)
	{
		if (ptry)
		{
			CantOrd *x_order = ptrx -> GetOrder ();
			CantOrd *y_order = ptry -> GetOrder ();
			if (*x_order == *y_order)
			{	// members of the same order
				result -> Append (CantOrd (V_OMEGA, x_order -> Copy (), NMinus (ptrx->mult, ptry->mult), NULL));
				if (ptrx -> next) result -> Append (*ptrx -> next);
			}
			else result -> Append (*ptrx);
			delete x_order;
			delete y_order;
		}
		else result -> Append (*ptrx);
	}

	result -> Renorm ();
	return result;
}


CantOrd *CantOrd::GetTimes (const CantOrd &x, const CantOrd &y) const
{
	CantOrd *result = CantOrd ().Copy ();
	const CantOrd *ptrx = &x, *ptry = &y;
	CantOrd *x_order = x.GetOrder ();
	int isinf = TRUE;

	// sum w^(a1+bj)*mj
	do
	{
		CantOrd *y_order = ptry -> GetOrder ();
		if (*y_order != CantOrd (0))
		{
			CantOrd exponent;
			exponent.Plus (*x_order, *y_order);
			result -> Append (CantOrd (V_OMEGA, exponent.Copy (), ptry -> mult, NULL));
			ptry = ptry -> next;
		}
		else isinf = FALSE;
		delete y_order;
	}
	while (isinf && ptry);

	long m = ptry ? ptry -> mult : 0;
	if (m != 0)
	{
		// + w^a1*n1*m + sum_{i=2..k} w^ai*mi
		result -> Append (CantOrd (V_OMEGA, x_order -> Copy (), NTimes (ptrx -> mult, m), NULL));
		result -> Append (*ptrx -> next);
	}

	result -> Renorm ();
	delete x_order;
	return result;
}


#pragma argsused
CantOrd *CantOrd::GetDiv (const CantOrd &x, const CantOrd &y) const
{
	SetOrdError (ORD_ERR_IMPL);
	return new CantOrd ();
}


#pragma argsused
CantOrd *CantOrd::GetMod (const CantOrd &x, const CantOrd &y) const
{
	SetOrdError (ORD_ERR_IMPL);
	return new CantOrd ();
}


CantOrd *CantOrd::GetPower (const CantOrd &x, const CantOrd &y) const
{
	CantOrd *result = CantOrd ().Copy ();
	CantOrd *x_order = x.GetOrder ();

	if (*x_order == CantOrd (0))  // n^y
	{
		assert (!x.next); // cannot have next

		if (x.mult == 0)
			if (y == CantOrd (0))  *result = CantOrd (1);
			else                   *result = CantOrd (0);
		else if (x.mult == 1)      *result = CantOrd (1);
		else
		{
			CantOrd exponent = CantOrd ();
			const CantOrd *ptry = &y;
			int isinf = TRUE;
			do // exponent construction
			{
				CantOrd *y_order = ptry -> GetOrder ();
				if (*y_order != CantOrd (0))
				{
					CantOrd eexponent;
					eexponent.Minus (*y_order, CantOrd (1));
					exponent.Append (CantOrd (V_OMEGA, eexponent.Copy (), ptry -> mult, NULL));
					ptry = ptry -> next;
				}
				else isinf = FALSE;
				delete y_order;
			}
			while (isinf && ptry);

			long m = ptry ? ptry -> mult : 0;
			*result = CantOrd (V_OMEGA, exponent.Copy (), NPower (x.mult, m), NULL);
		}
	}
	else // x^y
	{
		CantOrd infp, finp;

		// construct the first factor - from the infinite part of y
		CantOrd exponent = CantOrd ();
		const CantOrd *ptry = &y;
		int isinf = TRUE;
		do
		{
			CantOrd *y_order = ptry -> GetOrder ();
			if (*y_order != CantOrd (0))
			{
				exponent.Append (CantOrd (*ptry, 1));
				ptry = ptry -> next;
			}
			else isinf = FALSE;
			delete y_order;
		}
		while (isinf && ptry);
		exponent.Renorm ();
		exponent.Times (*x_order, exponent);
		infp = CantOrd (V_OMEGA, exponent.Copy (), 1, NULL);

		// construct the second factor - from the finite part of y
		long m = ptry ? ptry -> mult : 0;
		if (m == 0)      finp = CantOrd (1);
		else if (m == 1) finp <<= x;
		else
		{
			CantOrd exponent;
			exponent.Times (*x_order, CantOrd (m));
			finp = CantOrd (V_OMEGA, exponent.Copy (), x.mult, NULL);

			if (x.next)
			{
				int num;
				for (num = 1; --m >= 0 && num <= ORD_POW_MAX_MEMB;)
				{
					const CantOrd *ptrx = x.next;
					int isinf = TRUE;
					do
					{
						CantOrd *xi_order = ptrx -> GetOrder ();
						if (*xi_order != CantOrd (0))
						{
							if ((num += (ptrx -> mult != 0)) > ORD_POW_MAX_MEMB)
							{
								SetOrdError (ORD_ERR_TOO);
								break;
							}
							else
							{
								exponent.Times (*x_order, CantOrd (m));
								exponent += *xi_order;
								finp.Append (CantOrd (V_OMEGA, exponent.Copy (), ptrx -> mult, NULL));
								ptrx = ptrx -> next;
							}
						}
						else isinf = FALSE;
						delete xi_order;
					}
					while (isinf && ptrx);

					long n = ptrx ? ptrx -> mult : 0;
					if (n == 0) { m = 0; break; } // the other mult.rule

					if ((num += (n != 0)) > ORD_POW_MAX_MEMB)
					{
						SetOrdError (ORD_ERR_TOO);
						break;
					}
					else
					{
						exponent.Times (*x_order, CantOrd (m));
						finp.Append (CantOrd (V_OMEGA, exponent.Copy (), m ? NTimes (x.mult, n) : n, NULL));
					}
				} // for
			} // if (x.next)
		} // else

		// construct the result by multiplying both factors
		result -> Times (infp, finp);
	}

	result -> Renorm ();
	delete x_order;
	return result;
}


#undef SIGN

/********************************************************************/

#endif  // if __ORD
#endif  // ifndef __ORD
